@@ -58,7 +58,7 @@ class SegmentItem(pg.LinearRegionItem):
             self.bounds = time_bounds
         self.event_index = event_index
         self.xrange = xrange
-        # this corresponds to the undelrying only for TraveView, not for SpecView
+        # this corresponds to the underlying view only for TraceView, not for SpecView
         self._parent = self.getViewWidget
 
 
@@ -206,6 +206,13 @@ class TraceView(pg.PlotWidget):
         self.callback = callback
         self.getPlotItem().mouseClickEvent = self._click
 
+        self.threshold_line = pg.InfiniteLine(movable=True, angle=0,
+                                pos=0,
+                                pen=pg.mkPen(color='r', width=1),
+                                bounds=[0, None],
+                                label='Threshold',
+                                labelOpts={'position': 0.9})
+
     @property
     def m(self):  # make read-only
         return self._m
@@ -222,13 +229,18 @@ class TraceView(pg.PlotWidget):
     def yrange(self):
         return np.array(self.viewRange()[1])
 
+    @property
+    def threshold(self):
+        if self.threshold_line is not None:
+            return self.threshold_line.value()
+
     def update_trace(self):
         self.clear()
         if self.m.nb_channels is not None and self.m.show_all_channels and self.m.y_other is not None:
             for chan in range(self.m.nb_channels - 1):
                         self.addItem(pg.PlotCurveItem(self.m.x[::self.m.step],
                                                       self.m.y_other[::self.m.step, chan],
-                                                      pen=pg.mkPen(color=[128, 128, 128])))
+                                                      pen=pg.mkPen(color=[127, 127, 127])))
         # plot selected trace
         self.addItem(pg.PlotCurveItem(self.m.x[::self.m.step],
                                       np.array(self.m.y[::self.m.step])))
@@ -237,6 +249,11 @@ class TraceView(pg.PlotWidget):
         self.addItem(pg.InfiniteLine(movable=False, angle=90,
                                      pos=self.m.x[int(self.m.span / 2)],
                                      pen=pg.mkPen(color='r', width=1)))
+        if self.m.threshold_mode:
+            self.addItem(self.threshold_line)
+            self.addItem(pg.PlotCurveItem(self.m.x[::self.m.step],
+                                          self.m.envelope[::self.m.step],
+                                          pen=pg.mkPen(color=[255, 127, 127])))
 
     def add_segment(self, onset, offset, region_typeindex, brush=None, movable=True):
         region = SegmentItem((onset, offset), region_typeindex, self.xrange,
@@ -246,6 +263,8 @@ class TraceView(pg.PlotWidget):
             region.sigRegionChangeFinished.connect(self.m.on_region_change_finished)
 
     def add_event(self, xx, event_type, pen, movable=False):
+        if not len(xx):
+            return
         if not movable:
             xx = np.broadcast_to(xx[np.newaxis, :], (2, len(xx)))
             yy = np.zeros_like(xx) + self.yrange[:, np.newaxis]
@@ -374,6 +393,8 @@ class SpecView(pg.ImageView):
             region.sigRegionChangeFinished.connect(self.m.on_region_change_finished)
 
     def add_event(self, xx, event_type, pen, movable=False):
+        if not len(xx):
+            return
         xx0 = xx.copy()
         xx = self.time_to_pos(xx)
         if not movable:
