@@ -1,11 +1,13 @@
 """Load files created by various analysis programs created."""
 import numpy as np
+import pandas as pd
 import h5py
 import flammkuchen as dd_io
+from typing import Sequence, Optional
+import logging
 
 import scipy.interpolate
 import scipy.ndimage
-
 from scipy.io import loadmat
 import scipy.signal
 import scipy.stats
@@ -16,8 +18,9 @@ import xarray as xr
 import zarr
 import math
 from . import xarray_behave as xb
-from . import event_utils
-from typing import Sequence, Union
+from . import (event_utils,
+               annot)
+
 
 
 def rotate_point(pos, degrees, origin=(0, 0)):
@@ -166,9 +169,38 @@ def merge_channels(data, sampling_rate, filter_data: bool = True):
     return data_merged_max
 
 
+def fill_gaps(sine_pred, gap_dur=100):
+    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
+    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
+    if len(onsets) and len(offsets):
+        onsets = onsets[onsets<offsets[-1]]
+        offsets = offsets[offsets>onsets[0]]
+        durations = offsets - onsets
+        for idx, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
+            if idx>0 and offsets[idx-1]>onsets[idx]-gap_dur:
+                sine_pred[offsets[idx-1]:onsets[idx]+1] = 1
+    return sine_pred
+
+
+def remove_short(sine_pred, min_len=100):
+    # remove too short sine songs
+    onsets = np.where(np.diff(sine_pred.astype(np.int))==1)[0]
+    offsets = np.where(np.diff(sine_pred.astype(np.int))==-1)[0]
+    if len(onsets) and len(offsets):
+        onsets = onsets[onsets<offsets[-1]]
+        offsets = offsets[offsets>onsets[0]]
+        durations = offsets - onsets
+        for cnt, (onset, offset, duration) in enumerate(zip(onsets, offsets, durations)):
+            if duration<min_len:
+                sine_pred[onset:offset+1] = 0
+    return sine_pred
+
+
 def load_swap_indices(filepath):
     a = np.loadtxt(filepath, dtype=np.uintp)
     indices, flies1, flies2 = np.split(a, [1, 2], axis=-1)  # split columns in file into variables
+    if indices.ndim == 2:  # is sometimes 2d for some reason
+        indices = indices[:, 0]
     return indices, flies1, flies2
 
 
@@ -203,27 +235,6 @@ def swap_flies(dataset, indices, flies1=0, flies2=1):
     return dataset
 
 
-def infer_event_categories(data):
-    """Infer event type (event/segment) based on median interval between
-    event times.
-
-    Args:
-        data ([type]): binary matrix [samples x events]
-    """
-    event_categories = []
-    for evt in range(data.shape[1]):
-        if np.any(data[:, evt]):
-            dt = np.nanmedian(np.diff(np.where(data[:, evt])[0]))
-        else:
-            dt = 10  # fall back to events w/o annotations
-
-        if dt > 1:
-            event_categories.append('event')
-        else:
-            event_categories.append('segment')
-    return event_categories
-
-
 def load_segmentation_matlab(filepath):
     """Load output produced by FlySongSegmenter."""
     res = dict()
@@ -265,29 +276,51 @@ def load_segmentation(filepath):
     File should have at least 'event_names' and 'event_indices' datasets."""
     res = dd_io.load(filepath)
     # extract event_seconds
-    event_seconds = {}
-    for indices, name in zip(res['event_indices'], res['event_names']):
-        event_seconds[name] = indices / res['samplerate_Hz']
+    song_seconds = {}
+    song_categories = {}
 
-    # event_categories
-    if 'event_categories' in res:
-        event_categories = res['event_categories']
+    for event_indices, event_name in zip(res['event_indices'], res['event_names']):
+        song_seconds[event_name] = event_indices / res['samplerate_Hz']
+        song_categories[event_name] = 'event'
+
+    # post process segments ?
+    if 'sine' in res['segment_names']:
+        logging.info('   Found sine song - attempting to post process ')
+        sine_index = res['segment_names'].index('sine')
+        res['segment_labels'][sine_index] = fill_gaps(res['segment_labels'][sine_index],
+                                                        gap_dur = 20 / 1000 * res['samplerate_Hz'])
+        res['segment_labels'][sine_index] = remove_short(res['segment_labels'][sine_index],
+                                                            min_len = 20 / 1000 * res['samplerate_Hz'])
+
+    # extract segment on- and offsets
+    segment_indices = event_utils.traces_to_eventtimes(res['segment_labels'], res['segment_names'], ['segment'])
+    for segment_name, segment_indices in segment_indices.items():
+        song_seconds[segment_name] = segment_indices / res['samplerate_Hz']
+        song_categories[segment_name] = 'segment'
+
+    return song_seconds, song_categories
+
+
+def load_manual_annotation_csv(filepath):
+    """Load output produced by xb."""
+    df = pd.read_csv(filepath)
+
+    if not all([item in df.columns for item in ['name','start_seconds', 'stop_seconds']]):
+        logging.error(f"Malformed CSV file {filepath} - needs to have these columns: ['name','start_seconds', 'stop_seconds']. Returning empty results")
+        event_seconds = dict()
+        event_categories = dict()
     else:
-        event_categories = {}
-        for name, times in event_seconds.items():
-            if times.ndim == 1 or times.shape[1] == 1:
-                event_categories[name] = 'event'
-            else:
-                event_categories[name] = 'segment'
+        event_seconds = annot.Events.from_df(df)
+        event_categories = event_seconds.categories
     return event_seconds, event_categories
 
 
-def load_manual_annotation(filepath):
-    """Load output produced by the python ManualSegmenter."""
+def load_manual_annotation_zarr(filepath):
+    """Load output produced by xb (legacy format)."""
     manual_events_ds = xb.load(filepath)
 
     if 'event_categories' not in manual_events_ds:
-        event_categories_List = infer_event_categories(manual_events_ds.song_events.data)
+        event_categories_List = event_utils.infer_event_categories_from_traces(manual_events_ds.song_events.data)
 
         # force these to be the correct types even if they are empty and
         # the event_cat inference does not work
@@ -319,16 +352,16 @@ def load_manual_annotation_matlab(filepath):
             for key, val in f.items():
                 mat_data[key.lower()] = val[:].T
 
-    manual_events_seconds = dict()
+    events_seconds = dict()
     event_categories = dict()
     for key, val in mat_data.items():
         if len(val) and hasattr(val, 'ndim') and val.ndim == 2 and not key.startswith('_'):  # ignore matfile metadata
-            manual_events_seconds[key.lower() + '_manual'] = np.sort(val[:, 1:])
+            events_seconds[key.lower() + '_manual'] = np.sort(val[:, 1:])
             if val.shape[1] == 2:  # pulse times
                 event_categories[key.lower() + '_manual'] = 'event'
             else:  # sine on and offset
                 event_categories[key.lower() + '_manual'] = 'segment'
-    return manual_events_seconds, event_categories
+    return events_seconds, event_categories
 
 
 def load_tracks(filepath):
@@ -431,8 +464,8 @@ def load_poses_deepposekit(filepath):
         filepath (str): name of the file produced by DeepPoseKit
 
     Returns:
-        poses_ego (xarray.DataArray): poses in EGOcentric (centered around thorax, head aligned straight upwards
-        poses_allo (xarray.DataArray): poses in ALLOcentric (frame) coordinates
+        poses_ego (xarray.DataArray): poses in EGOcentric (centered around thorax, head aligned straight upwards), [frames, flies, bodypart, x/y]
+        poses_allo (xarray.DataArray): poses in ALLOcentric (frame) coordinates, [frames, flies, bodypart, x/y]
         partnames (List[str]): list of names for each body part (is already part of the poses_ego/allo xrs)
         first_pose_frame, last_pose_frame (int): frame corresponding to first and last item in the poses_ego/allo arrays (could attach this info as xr dim)
     """
@@ -443,8 +476,8 @@ def load_poses_deepposekit(filepath):
 
     poses_allo = ds.poses + ds.box_centers - box_size/2
 
-    first_pose_frame = int(np.argmin(np.isnan(ds.poses[:, 0, 0, 0])))
-    last_pose_frame = int(np.argmin(~np.isnan(ds.poses[first_pose_frame:, 0, 0, 0])) + first_pose_frame)
+    first_pose_frame = int(np.argmin(np.isnan(ds.poses.data[:, 0, 0, 0]).data))
+    last_pose_frame = int(np.argmin(~np.isnan(np.array(ds.poses.data[first_pose_frame:, 0, 0, 0]).data)) + first_pose_frame)
     if last_pose_frame == 0:
         last_pose_frame = ds.poses.shape[0]
 
@@ -465,7 +498,18 @@ def load_poses_deepposekit(filepath):
     return poses_ego, poses_allo, ds.poseparts, first_pose_frame, last_pose_frame
 
 
-def load_raw_song(filepath_daq: str, song_channels: Union[Sequence[int], None] = None,
+def load_poses_sleap(filepath):
+    with h5py.File(filepath, 'r') as f:
+        pose_parts = f['node_names']
+        track_names = f['track_names']
+        track_occupancy = f['track_occupancy'][:]
+        tracks = f['tracks'][:]
+
+    poses_allo = tracks.transpose([3, 0, 2, 1])
+    return poses_ego, poses_allo, ds.poseparts, first_pose_frame, last_pose_frame
+
+
+def load_raw_song(filepath_daq: str, song_channels: Optional[Sequence[int]] = None,
                   return_nonsong_channels: bool = False, lazy: bool = False):
     """[summary]
 
