@@ -7,15 +7,18 @@ import zarr
 import logging
 import os.path
 from pathlib import Path
-from . import loaders as ld
-from . import metrics as mt
-from . import event_utils
+import pandas as pd
+from typing import List, Optional
+from . import (loaders as ld,
+               metrics as mt,
+               event_utils,
+               annot)
 
 
 def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_rate=1_000,
              keep_multi_channel: bool = False, resample_video_data: bool = True,
              include_song: bool = True, include_tracks: bool = True, include_poses: bool = True,
-             make_mask: bool = False, fix_fly_indices: bool = True) -> xr.Dataset:
+             fix_fly_indices: bool = True, pixel_size_mm: Optional[float] = None) -> xr.Dataset:
     """[summary]
 
     Args:
@@ -29,9 +32,8 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
         include_song (bool, optional): [description]. Defaults to True.
         include_tracks (bool, optional): [description]. Defaults to True.
         include_poses (bool, optional): [description]. Defaults to True.
-        make_mask (bool, optional): ..., Defaults to False.
         fix_fly_indices (bool, optional): Will attempt to load swap info and fix fly id's accordingly, Defaults to True.
-
+        pixel_size_mm (float, optional): Size of a pixel (in mm) in the video. Used to convert tracking data to mm.
     Returns:
         xarray.Dataset
     """
@@ -103,7 +105,6 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
 
     # instead of these flags - just check for existence of dict keys??
     with_segmentation = False
-    with_segmentation_matlab = False
     with_segmentation_manual = False
     with_segmentation_manual_matlab = False
     with_song = False
@@ -120,18 +121,23 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
         filepath_segmentation_matlab = Path(root, res_path, datename, f'{datename}_song.mat')
         filepath_segmentation = Path(root, res_path, datename, f'{datename}_song.h5')
         if os.path.exists(filepath_segmentation):
-            # res = ld.load_segmentation(filepath_segmentation)
-            auto_event_seconds, auto_event_categories = ld.load_segmentation(filepath_segmentation)
-            with_segmentation = True
-            merge_raw_recording = True
-            logging.debug(f'   {filepath_segmentation} loaded.')
+            try:
+                auto_event_seconds, auto_event_categories = ld.load_segmentation(filepath_segmentation)
+                with_segmentation = True
+                merge_raw_recording = False  # True
+                logging.debug(f'   {filepath_segmentation} loaded.')
+            except Exception as e:
+                logging.debug(f'   {filepath_segmentation} failed.')
+                logging.exception(e)
+                with_segmentation = False
+                merge_raw_recording = False
+
         elif os.path.exists(filepath_segmentation_matlab):
             logging.debug(f'   {filepath_segmentation} not found.')
-            # res = ld.load_segmentation_matlab(filepath_segmentation_matlab)
             auto_event_seconds, auto_event_categories  = ld.load_segmentation_matlab(filepath_segmentation_matlab)
             with_segmentation = True
             with_segmentation_matlab = True
-            merge_raw_recording = True
+            merge_raw_recording = False  # True
             with_song = True
             logging.debug(f'   {filepath_segmentation_matlab} loaded.')
         else:
@@ -141,19 +147,32 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
         # load MANUAL SONG ANNOTATIONS
         # first try PYTHON, then matlab
         logging.info('Attempting to load manual segmentation:')
-        filepath_segmentation_manual = Path(root, res_path, datename, f'{datename}_songmanual.zarr')
+
+        filepath_segmentation_manual_csv_new = Path(root, res_path, datename, f'{datename}_annotations.csv')
+        filepath_segmentation_manual_csv_old = Path(root, res_path, datename, f'{datename}_songmanual.csv')
+        if os.path.exists(filepath_segmentation_manual_csv_new):
+            filepath_segmentation_manual_csv = filepath_segmentation_manual_csv_new
+        else:
+            filepath_segmentation_manual_csv = filepath_segmentation_manual_csv_old
+
+        filepath_segmentation_manual_zarr = Path(root, res_path, datename, f'{datename}_songmanual.zarr')
         filepath_segmentation_manual_matlab = Path(root, res_path, datename, f'{datename}_songmanual.mat')
-        if os.path.exists(filepath_segmentation_manual):
-            manual_event_seconds, manual_event_categories = ld.load_manual_annotation(filepath_segmentation_manual)  # need to extract event_seconds and categories from that one
+
+        if os.path.exists(filepath_segmentation_manual_csv):
+            manual_event_seconds, manual_event_categories = ld.load_manual_annotation_csv(filepath_segmentation_manual_csv)  # need to extract event_seconds and categories from that one
             with_segmentation_manual = True
-            logging.debug(f'   {filepath_segmentation_manual} loaded.')
+            logging.debug(f'   {filepath_segmentation_manual_csv} loaded.')
+        elif os.path.exists(filepath_segmentation_manual_zarr):
+            manual_event_seconds, manual_event_categories = ld.load_manual_annotation_zarr(filepath_segmentation_manual_zarr)  # need to extract event_seconds and categories from that one
+            with_segmentation_manual = True
+            logging.debug(f'   {filepath_segmentation_manual_zarr} loaded.')
         elif os.path.exists(filepath_segmentation_manual_matlab):
             manual_event_seconds, manual_event_categories = ld.load_manual_annotation_matlab(filepath_segmentation_manual_matlab)
             with_segmentation_manual = True
             with_segmentation_manual_matlab = True
             logging.debug(f'   {filepath_segmentation_manual_matlab} loaded.')
         else:
-            logging.info(f'Could not load manual segmentation from {filepath_segmentation_manual} or {filepath_segmentation_manual_matlab}.')
+            logging.info(f'Could not load manual segmentation from {filepath_segmentation_manual_csv} or {filepath_segmentation_manual_zarr} or {filepath_segmentation_manual_matlab}.')
 
         # load RAW song traces
         res = dict()
@@ -209,7 +228,7 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
 
     # PREPARE DataArrays
     dataset_data = dict()
-    if with_song:  # MERGED song recording
+    if with_song and merge_raw_recording:  # MERGED song recording
         song = xr.DataArray(data=res['song'][first_sample:last_sample, 0],  # cut recording to match new grid
                             dims=['sampletime'],
                             coords={'sampletime': sampletime, },
@@ -280,9 +299,20 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
         song_events_ds = event_utils.eventtimes_to_traces(song_events.to_dataset(name='song_events'),
                                                         song_events.attrs['event_times'])
         dataset_data['song_events'] = song_events_ds.song_events
+        try:
+            ds_eventtimes = annot.Events(event_seconds).to_dataset()
+            dataset_data['event_times'] = ds_eventtimes.event_times
+            dataset_data['event_names'] = ds_eventtimes.event_names
+            dataset_data['event_names'].attrs['possible_event_names'] = ds_eventtimes.attrs['possible_event_names']
+        except Exception as e:
+            logging.error('Failed to generate event_times data arrays:')
+            logging.exception(e)
 
     # BODY POSITION
     fps = None
+    if pixel_size_mm is None:
+        pixel_size_mm = np.nan
+
     if with_tracks:
         logging.info('   tracking')
         frame_numbers = np.arange(first_tracked_frame, last_tracked_frame)
@@ -314,6 +344,7 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
                                         'time_units': 'seconds',
                                         'video_fps': fps,
                                         'spatial_units': 'pixels',
+                                        'pixel_size_mm': pixel_size_mm,
                                         'background': background,
                                         'tracks_fixed': with_fixed_tracks})
         dataset_data['body_positions'] = positions
@@ -355,6 +386,7 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
                                     'time_units': 'seconds',
                                     'video_fps': fps,
                                     'spatial_units': 'pixels',
+                                    'pixel_size_mm': pixel_size_mm,
                                     'poses_from': poses_from})
         dataset_data['pose_positions'] = poses
 
@@ -370,6 +402,7 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
                                          'time_units': 'seconds',
                                          'video_fps': fps,
                                          'spatial_units': 'pixels',
+                                         'pixel_size_mm': pixel_size_mm,
                                          'poses_from': poses_from})
         dataset_data['pose_positions_allo'] = poses_allo
     # MAKE THE DATASET
@@ -380,6 +413,10 @@ def assemble(datename, root='', dat_path='dat', res_path='res', target_sampling_
     if 'nearest_frame' not in dataset:
         dataset.coords['nearest_frame'] = (('time'), nearest_frame)
 
+    # convert spatial units to mm using info in attrs
+    dataset = convert_spatial_units(dataset, to_units='mm',
+                                    names = ['body_positions', 'pose_positions', 'pose_positions_allo'])
+
     if target_sampling_rate is None:
         target_sampling_rate = fps
     # save command line args
@@ -539,6 +576,41 @@ def assemble_metrics(dataset, make_abs: bool = True, make_rel: bool = True, smoo
     return feature_dataset
 
 
+def convert_spatial_units(ds: xr.Dataset, to_units: Optional[str] = None, names: Optional[List[str]] = None) -> xr.Dataset:
+    """px <-> mm using attrs['pixels_to_mm'] and attrs['spatial_units'].
+
+    Args:
+        ds (xr.Dataset): xb Dataset to process
+        to_units (Optional[str], optional): Target units - either 'pixels' or 'mm'. Defaults to None (always convert mm <-> px).
+        names (Optional[List[str]], optional): Names of DataArrays/sets in ds to convert. Defaults to None.
+
+    Returns:
+        xr.Dataset: with converted spatial units
+    """
+    if names is None:
+        names = ['body_positions', 'pose_positions', 'pose_positions_allo']
+
+    for name in names:
+        if name in ds:
+            da = ds[name]
+            spatial_units = da.attrs['spatial_units']
+            pixel_size_mm = da.attrs['pixel_size_mm']
+
+            if pixel_size_mm is np.nan:  # can't convert so skip
+                continue
+
+            if to_units is not None and spatial_units == to_units:  # already in the correct units so skip
+                continue
+
+            if spatial_units == 'mm':  # convert to pixels
+                da /= pixel_size_mm
+                da.attrs['spatial_units'] = 'pixels'
+            elif spatial_units == 'pixels':  # convert to mm
+                da *= pixel_size_mm
+                da.attrs['spatial_units'] = 'mm'
+    return ds
+
+
 def save(savepath, dataset):
     """[summary]
 
@@ -554,21 +626,21 @@ def save(savepath, dataset):
             del dataset.song_events.attrs['event_times']
 
     with zarr.ZipStore(savepath, mode='w') as zarr_store:
-        ## re-chunking does not seem to help with IO speed upon lazy loading
-        # chunks = dict(dataset.dims)
-        # chunks['time'] = 100_000
-        # chunks['sampletime'] = 100_000
-        # dataset = dataset.chunk(chunks)
+        # re-chunking does not seem to help with IO speed upon lazy loading
+        chunks = dict(dataset.dims)
+        chunks['time'] = 100_000
+        chunks['sampletime'] = 100_000
+        dataset = dataset.chunk(chunks)
         dataset.to_zarr(store=zarr_store, compute=True)
 
 
 def _normalize_strings(dataset):
     """Ensure all keys in coords are proper (unicode?) python strings, not byte strings."""
     dn = dict()
+
     for key, val in dataset.coords.items():
         if val.dtype == 'S16':
-            val.data = np.array([v.decode() for v in val.data])
-        dn[key] = val
+            dataset[key] = [v.decode() for v in val.data]
     return dataset
 
 
@@ -598,11 +670,75 @@ def load(savepath, lazy: bool = False, normalize_strings: bool = True,
 
     if normalize_strings:
         dataset = _normalize_strings(dataset)
-
+    print(dataset)
     return dataset
 
+def load_npz(filepath: str, dataset: Optional[str] = 'song'):
+    import numpy as np
+    logging.info(f"Loading data from NPZ file at {filepath}.")
+    with np.load(filepath) as file:
+        try:
+            sampling_rate = file['samplerate']
+        except KeyError:
+            sampling_rate = None
+        data = file[dataset]
+    return data, sampling_rate
 
-def from_wav(filepath, target_samplerate=None, event_names=[], event_categories=[]):
+def load_npy(filepath: str, dataset: Optional[str] = None):
+    import numpy as np
+    logging.info(f"Loading data from NPY file {filepath}.")
+    data = np.load(filepath)
+    samplingrate = None
+    return data, samplingrate
+
+def load_audio(filepath: str, dataset: Optional[str] = None):
+    import soundfile
+    logging.info(f"Loading data from audio file at {filepath}.")
+    data, sampling_rate = soundfile.read(filepath)
+    return data, sampling_rate
+
+def load_wav(filepath: str, dataset: Optional[str] = None):
+    import scipy.io.wavfile
+    samplerate, data = scipy.io.wavfile.read(filepath)
+    data = data[:, np.newaxis] if data.ndim==1 else data  # adds singleton dim for single-channel wavs
+    return data, samplerate
+
+def load_h5(filepath: str, dataset: Optional[str] = 'samples'):
+    import h5py
+    logging.info(f"Loading dataset {dataset} from HDF5 file {filepath}.")
+    with h5py.File(filepath, 'r') as f:
+        data = f[dataset][:]
+    samplingrate = None
+    return data, samplingrate
+
+# def zarr(filepath):
+#     import zarr
+#     logging.info(f"Loading zarr file {filepath}.")
+#     xb.load()
+#     samplingrate = None
+#     return data, samplingrate
+
+data_loaders = {'audio': load_audio, 'npy': load_npy, 'npz': load_npz, 'h5': load_h5, 'wav': load_wav}
+
+def from_file(filepath: str, loader_name: str = 'audio', target_samplerate: Optional[float] = None,
+              samplerate: Optional[float] = None, dataset: Optional[str] = None,
+              event_names=[], event_categories=[], annotation_path: Optional[str] = None):
+    # TODO merge with from_data
+
+    data, samplerate_from_file = data_loaders[loader_name](filepath, dataset)
+
+    if samplerate is None:
+        samplerate = samplerate_from_file
+
+    if data.ndim==1:
+        logging.info("   Data is 1d so prolly single-channel audio - appending singleton dimension.")
+        data = data[:, np.newaxis]
+
+    if event_names is None:
+        event_names = []
+
+    if event_categories is None:
+        event_categories = []
 
     if event_names and not event_categories:
         logging.info('No event_categories specified - defaulting to segments')
@@ -611,52 +747,49 @@ def from_wav(filepath, target_samplerate=None, event_names=[], event_categories=
     if len(event_names) != len(event_categories):
         raise ValueError(f'event_names and event_categories need to have same length - have {len(event_names)} and {len(event_categories)}.')
 
-    import soundfile
-    data, sampling_rate = soundfile.read(filepath)
-
     if target_samplerate is None:
-        target_samplerate = sampling_rate
+        target_samplerate = samplerate
 
     dataset_data = dict()
 
-    data = data[:, np.newaxis]
-    sampletime = np.arange(len(data)) / sampling_rate
+    sampletime = np.arange(len(data)) / samplerate
     time = np.arange(sampletime[0], sampletime[-1], 1 / target_samplerate)
 
     song_raw = xr.DataArray(data=data,  # cut recording to match new grid
                         dims=['sampletime', 'channels'],
                         coords={'sampletime': sampletime, },
                         attrs={'description': 'Song signal merged across all recording channels.',
-                               'sampling_rate_Hz': sampling_rate,
+                               'sampling_rate_Hz': samplerate,
                                'time_units': 'seconds',
                                'amplitude_units': 'volts'})
     dataset_data['song_raw'] = song_raw
 
+    event_times = annot.Events()
+    if annotation_path is not None and os.path.exists(annotation_path):
+        try:
+            df = pd.read_csv(annotation_path)
+            event_times = annot.Events.from_df(df)
+        except Exception as e:
+            logging.exception(e)
+
     if event_names is not None and event_categories is not None:
-        song_events_data = np.zeros((len(time), len(event_names)), dtype=np.uint)
-        song_events = xr.DataArray(data=song_events_data,
-                    dims=['time', 'event_types'],
-                    coords={'time': time,
-                            'event_types': event_names,
-                            'event_categories': (('event_types'), event_categories)},
-                    attrs={'description': 'Song annotations',
-                            'sampling_rate_Hz': sampling_rate,
-                            'time_units': 'seconds',
-                           })
-        dataset_data['song_events'] = song_events
+        for event_name, event_category in zip(event_names, event_categories):
+            if event_name not in event_times:
+                event_times.add_name(event_name, event_category)
+
+    ds_eventtimes = event_times.to_dataset()
+    dataset_data['event_names'] = ds_eventtimes.event_names
+    dataset_data['event_times'] = ds_eventtimes.event_times
 
     # MAKE THE DATASET
     ds = xr.Dataset(dataset_data, attrs={})
     ds.coords['time'] = time
-    ds.coords['nearest_frame'] = ('time', (time/100).astype(np.uint))  # do we need this?
 
     # save command line args
     ds.attrs = {'video_filename': '',
                 'datename': filepath,
                 'root': '', 'dat_path': '', 'res_path': '',
-                'sampling_rate_Hz': sampling_rate,
-                     'sampling_rate_Hz': sampling_rate,
-                'sampling_rate_Hz': sampling_rate,
+                'sampling_rate_Hz': samplerate,
                 'target_sampling_rate_Hz': target_samplerate}
     logging.info(ds)
     return ds
